@@ -39,8 +39,7 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
 
     hdr_prefix = 'X-Object-'
     known_options = {'no-ssl', 'ssl-ca-path', 'tcp-timeout',
-                     'disable-expect100', 'no-feature-detection',
-                     'domain', 'project_domain'}
+                     'disable-expect100', 'no-feature-detection'}
 
     _add_meta_headers = s3c.Backend._add_meta_headers
     _extractmeta = s3c.Backend._extractmeta
@@ -475,7 +474,7 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
         resp = self._do_request('POST', '/', subres='bulk-delete', body=body, headers=headers)
 
         # bulk deletes should always return 200
-        if resp.status is not 200:
+        if resp.status != 200:
             raise HTTPError(resp.status, resp.reason, resp.headers)
 
         hit = re.match(r'^application/json(;\s*charset="?(.+?)"?)?$',
@@ -497,7 +496,7 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
         except ValueError:
             raise RuntimeError('Unexpected server reply')
 
-        if resp_status_code is 200:
+        if resp_status_code == 200:
             # No errors occured, everything has been deleted
             del keys[:]
             return
@@ -762,7 +761,7 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
                           self._dump_response(resp, body=conn.read(2048)))
                 raise HTTPError(resp.status, resp.reason, resp.headers)
 
-            if resp.status is 200:
+            if resp.status == 200:
                 hit = re.match(r'^application/json(;\s*charset="?(.+?)"?)?$',
                 resp.headers['content-type'])
                 if not hit:
