@@ -12,17 +12,22 @@ from .backends.common import NoSuchObject
 from .multi_lock import MultiLock
 from .logging import logging # Ensure use of custom logger class
 from collections import OrderedDict
-from contextlib import contextmanager
-from llfuse import lock, lock_released
 from queue import Queue, Empty as QueueEmpty, Full as QueueFull
+from argparse import Namespace
 import os
 import hashlib
 import shutil
 import threading
 import time
+import trio
 import re
 import sys
 
+try:
+    from contextlib import asynccontextmanager
+except ModuleNotFoundError:
+    from async_generator import asynccontextmanager
+
 # standard logger for this module
 log = logging.getLogger(__name__)
 
@@ -41,103 +46,6 @@ class NoWorkerThreads(Exception):
 
     pass
 
-class Distributor(object):
-    '''
-    Distributes objects to consumers.
-    '''
-
-    def __init__(self):
-        super().__init__()
-
-        self.slot = None
-        self.cv = threading.Condition()
-
-        #: Number of threads waiting to consume an object
-        self.readers = 0
-
-    def put(self, obj, timeout=None):
-        '''Offer *obj* for consumption
-
-        The method blocks until another thread calls `get()` to consume the
-        object.
-
-        Return `True` if the object was consumed, and `False` if *timeout* was
-        exceeded without any activity in the queue (this means an individual
-        invocation may wait for longer than *timeout* if objects from other
-        threads are being consumed).
-        '''
-
-        if obj is None:
-            raise ValueError("Can't put None into Queue")
-
-        with self.cv:
-            # Wait until a thread is ready to read
-            while self.readers == 0 or self.slot is not None:
-                log.debug('waiting for reader..')
-                if not self.cv.wait(timeout):
-                    log.debug('timeout, returning')
-                    return False
-
-            log.debug('got reader, enqueueing %s', obj)
-            self.readers -= 1
-            assert self.slot is None
-            self.slot = obj
-            self.cv.notify_all() # notify readers
-
-        return True
-
-    def get(self):
-        '''Consume and return an object
-
-        The method blocks until another thread offers an object by calling the
-        `put` method.
-        '''
-        with self.cv:
-            self.readers += 1
-            self.cv.notify_all()
-            while self.slot is None:
-                log.debug('waiting for writer..')
-                self.cv.wait()
-            tmp = self.slot
-            self.slot = None
-            self.cv.notify_all()
-
-        return tmp
-
-
-class SimpleEvent(object):
-    '''
-    Like threading.Event, but without any internal flag. Calls
-    to `wait` always block until some other thread calls
-    `notify` or `notify_all`.
-    '''
-
-    def __init__(self):
-        super().__init__()
-        self.__cond = threading.Condition(threading.Lock())
-
-    def notify_all(self):
-        self.__cond.acquire()
-        try:
-            self.__cond.notify_all()
-        finally:
-            self.__cond.release()
-
-    def notify(self):
-        self.__cond.acquire()
-        try:
-            self.__cond.notify()
-        finally:
-            self.__cond.release()
-
-    def wait(self, timeout=None):
-        self.__cond.acquire()
-        try:
-            return self.__cond.wait(timeout)
-        finally:
-            self.__cond.release()
-
-
 class CacheEntry(object):
     """An element in the block cache
 
@@ -163,7 +71,9 @@ class CacheEntry(object):
         self.blockno = blockno
         self.last_write = 0
         self.pos = self.fh.tell()
-        self.size = os.fstat(self.fh.fileno()).st_size
+        # use allocation size instead of st_size
+        # to properly account for small files
+        self.size = os.fstat(self.fh.fileno()).st_blocks * 512
 
     def read(self, size=None):
         buf = self.fh.read(size)
@@ -244,9 +154,6 @@ class BlockCache(object):
     This class manages access to file blocks. It takes care of creation,
     uploading, downloading and deduplication.
 
-    This class uses the llfuse global lock. Methods which release the lock have
-    are marked as such in their docstring.
-
     Attributes
     ----------
 
@@ -274,7 +181,7 @@ class BlockCache(object):
         self.in_transit = set()
         self.upload_threads = []
         self.removal_threads = []
-        self.transfer_completed = SimpleEvent()
+        self.transfer_completed = trio.Condition()
 
         # Will be initialized once threads are available
         self.to_upload = None
@@ -286,8 +193,9 @@ class BlockCache(object):
         else:
             os.mkdir(self.path)
 
-        # Initialized fromt the outside to prevent cyclic dependency
-        self.fs = None
+        # Initialized fromt the outside to prevent cyclic dependency,
+        # used only to set failsafe attribute
+        self.fs = Namespace()
 
     def load_cache(self):
         '''Initialize cache from disk'''
@@ -311,7 +219,8 @@ class BlockCache(object):
     def init(self, threads=1):
         '''Start worker threads'''
 
-        self.to_upload = Distributor()
+        self.trio_token = trio.lowlevel.current_trio_token()
+        self.to_upload = trio.open_memory_channel(0)
         for _ in range(threads):
             t = threading.Thread(target=self._upload_loop)
             t.start()
@@ -333,59 +242,18 @@ class BlockCache(object):
                 t.start()
                 self.removal_threads.append(t)
 
-    def _lock_obj(self, obj_id, release_global=False):
-        '''Acquire lock on *obj*id*'''
-
-        if release_global:
-            with lock_released:
-                self.mlock.acquire(obj_id)
-        else:
-            self.mlock.acquire(obj_id)
-
-    def _unlock_obj(self, obj_id, release_global=False, noerror=False):
-        '''Release lock on *obj*id*'''
-
-        if release_global:
-            with lock_released:
-                self.mlock.release(obj_id, noerror=noerror)
-        else:
-            self.mlock.release(obj_id, noerror=noerror)
-
-    def _lock_entry(self, inode, blockno, release_global=False, timeout=None):
-        '''Acquire lock on cache entry'''
-
-        if release_global:
-            with lock_released:
-                return self.mlock.acquire((inode, blockno), timeout=timeout)
-        else:
-            return self.mlock.acquire((inode, blockno), timeout=timeout)
-
-    def _unlock_entry(self, inode, blockno, release_global=False,
-                      noerror=False):
-        '''Release lock on cache entry'''
-
-        if release_global:
-            with lock_released:
-                self.mlock.release((inode, blockno), noerror=noerror)
-        else:
-            self.mlock.release((inode, blockno), noerror=noerror)
-
-    def destroy(self, keep_cache=False):
-        '''Clean up and stop worker threads
-
-        This method should be called without the global lock held.
-        '''
+    async def destroy(self, keep_cache=False):
+        '''Clean up and stop worker threads'''
 
         log.debug('Flushing cache...')
         try:
-            with lock:
-                if keep_cache:
-                    self.flush() # releases global lock
-                    for el in self.cache.values():
-                        assert not el.dirty
-                        el.close()
-                else:
-                    self.drop()
+            if keep_cache:
+                await self.flush()
+                for el in self.cache.values():
+                    assert not el.dirty
+                    el.close()
+            else:
+                await self.drop()
         except NoWorkerThreads:
             log.error('Unable to flush cache, no upload threads left alive')
 
@@ -394,7 +262,7 @@ class BlockCache(object):
         log.debug('Signaling upload threads...')
         try:
             for t in self.upload_threads:
-                self._queue_upload(QuitSentinel)
+                await self._queue_upload(QuitSentinel)
         except NoWorkerThreads:
             pass
 
@@ -407,7 +275,7 @@ class BlockCache(object):
 
         log.debug('waiting for upload threads...')
         for t in self.upload_threads:
-            t.join()
+            await trio.to_thread.run_sync(t.join)
 
         log.debug('waiting for removal threads...')
         for t in self.removal_threads:
@@ -435,19 +303,29 @@ class BlockCache(object):
 
 
     def _upload_loop(self):
-        '''Process upload queue'''
+        '''Process upload queue.
+
+        This method runs in a separate thread outside the trio event loop.
+        '''
 
         while True:
-            tmp = self.to_upload.get()
+            log.debug('reading from upload queue...')
+            tmp = trio.from_thread.run(
+                self.to_upload[1].receive, trio_token=self.trio_token)
 
             if tmp is QuitSentinel:
+                log.debug('got QuitSentinel')
                 break
+            log.debug('got work')
 
             self._do_upload(*tmp)
 
 
     def _do_upload(self, el, obj_id):
-        '''Upload object'''
+        '''Upload object.
+
+        This method runs in a separate thread outside the trio event loop.
+        '''
 
         def do_write(fh):
             el.seek(0)
@@ -458,6 +336,42 @@ class BlockCache(object):
                 fh.write(buf)
             return fh
 
+        success = False
+        async def with_event_loop(exc_info):
+            if success:
+                self.db.execute('UPDATE objects SET size=? WHERE id=?', (obj_size, obj_id))
+                el.dirty = False
+            else:
+                exc = exc_info[1]
+                log.debug('upload of %d failed (%s: %s)', obj_id, type(exc).__name__, exc)
+                # At this point we have to remove references to this storage object
+                # from the objects and blocks table to prevent future cache elements
+                # to be de-duplicated against this (missing) one. However, this may
+                # already have happened during the attempted upload. The only way to
+                # avoid this problem is to insert the hash into the blocks table
+                # *after* successfull upload. But this would open a window without
+                # de-duplication just to handle the special case of an upload
+                # failing.
+                #
+                # On the other hand, we also want to prevent future deduplication
+                # against this block: otherwise the next attempt to upload the same
+                # cache element (by a different upload thread that has not
+                # encountered problems yet) is guaranteed to link against the
+                # non-existing block, and the data will be lost.
+                #
+                # Therefore, we just set the hash of the missing block to NULL,
+                # and rely on fsck to pick up the pieces. Note that we cannot
+                # delete the row from the blocks table, because the id will get
+                # assigned to a new block, so the inode_blocks entries will
+                # refer to incorrect data.
+                #
+                self.db.execute('UPDATE blocks SET hash=NULL WHERE obj_id=?', (obj_id,))
+
+            await self.mlock.release(obj_id)
+            await self.mlock.release(el.inode, el.blockno)
+            async with self.transfer_completed:
+                self.transfer_completed.notify_all()
+
         try:
             with self.backend_pool() as backend:
                 if log.isEnabledFor(logging.DEBUG):
@@ -471,52 +385,16 @@ class BlockCache(object):
                 else:
                     obj_size = backend.perform_write(do_write, 's3ql_data_%d'
                                                      % obj_id).get_obj_size()
-
-            with lock:
-                self.db.execute('UPDATE objects SET size=? WHERE id=?', (obj_size, obj_id))
-                el.dirty = False
-
-        except Exception as exc:
-            log.debug('upload of %d failed: %s', obj_id, exc)
-            # At this point we have to remove references to this storage object
-            # from the objects and blocks table to prevent future cache elements
-            # to be de-duplicated against this (missing) one. However, this may
-            # already have happened during the attempted upload. The only way to
-            # avoid this problem is to insert the hash into the blocks table
-            # *after* successfull upload. But this would open a window without
-            # de-duplication just to handle the special case of an upload
-            # failing.
-            #
-            # On the other hand, we also want to prevent future deduplication
-            # against this block: otherwise the next attempt to upload the same
-            # cache element (by a different upload thread that has not
-            # encountered problems yet) is guaranteed to link against the
-            # non-existing block, and the data will be lost.
-            #
-            # Therefore, we just set the hash of the missing block to NULL,
-            # and rely on fsck to pick up the pieces. Note that we cannot
-            # delete the row from the blocks table, because the id will get
-            # assigned to a new block, so the inode_blocks entries will
-            # refer to incorrect data.
-            #
-
-            with lock:
-                self.db.execute('UPDATE blocks SET hash=NULL WHERE obj_id=?',
-                                (obj_id,))
-            raise
-
+            success = True
         finally:
             self.in_transit.remove(el)
-            self._unlock_obj(obj_id)
-            self._unlock_entry(el.inode, el.blockno)
-            self.transfer_completed.notify_all()
+            trio.from_thread.run(with_event_loop, sys.exc_info(), trio_token=self.trio_token)
 
 
-    def wait(self):
+    async def wait(self):
         '''Wait until an object has been uploaded
 
-        If there are no objects in transit, return immediately. This method
-        releases the global lock.
+        If there are no objects in transit, return immediately.
         '''
 
         # Loop to avoid the race condition of a transfer terminating
@@ -525,15 +403,15 @@ class BlockCache(object):
             if not self.transfer_in_progress():
                 return
 
-            with lock_released:
-                if self.transfer_completed.wait(timeout=5):
+            with trio.move_on_after(5):
+                async with self.transfer_completed:
+                    await self.transfer_completed.wait()
                     return
 
-    def upload_if_dirty(self, el):
+    async def upload_if_dirty(self, el):
         '''Upload cache entry asynchronously
 
-        This method releases the global lock. Return True if the object
-        is actually scheduled for upload.
+        Return True if the object is actually scheduled for upload.
         '''
 
         log.debug('started with %s', el)
@@ -544,39 +422,42 @@ class BlockCache(object):
             return False
 
         # Calculate checksum
-        with lock_released:
-            self._lock_entry(el.inode, el.blockno)
-            added_to_transit = False
-            try:
-                if el is not self.cache.get((el.inode, el.blockno), None):
-                    log.debug('%s got removed while waiting for lock', el)
-                    self._unlock_entry(el.inode, el.blockno)
-                    return False
-                if el in self.in_transit:
-                    log.debug('%s already in transit', el)
-                    self._unlock_entry(el.inode, el.blockno)
-                    return True
-                if not el.dirty:
-                    log.debug('no longer dirty, returning')
-                    self._unlock_entry(el.inode, el.blockno)
-                    return False
+        await self.mlock.acquire(el.inode, el.blockno)
+        added_to_transit = False
+        try:
+            if el is not self.cache.get((el.inode, el.blockno), None):
+                log.debug('%s got removed while waiting for lock', el)
+                await self.mlock.release(el.inode, el.blockno)
+                return False
+            if el in self.in_transit:
+                log.debug('%s already in transit', el)
+                await self.mlock.release(el.inode, el.blockno)
+                return True
+            if not el.dirty:
+                log.debug('no longer dirty, returning')
+                await self.mlock.release(el.inode, el.blockno)
+                return False
 
-                log.debug('uploading %s..', el)
-                self.in_transit.add(el)
-                added_to_transit = True
-                sha = hashlib.sha256()
-                el.seek(0)
+            log.debug('uploading %s..', el)
+            self.in_transit.add(el)
+            added_to_transit = True
+            sha = hashlib.sha256()
+            el.seek(0)
+
+            def with_lock_released():
                 while True:
                     buf = el.read(BUFSIZE)
                     if not buf:
                         break
                     sha.update(buf)
-                hash_ = sha.digest()
-            except:
-                if added_to_transit:
-                    self.in_transit.discard(el)
-                self._unlock_entry(el.inode, el.blockno)
-                raise
+                return sha.digest()
+            hash_ = await trio.to_thread.run_sync(with_lock_released)
+
+        except:
+            if added_to_transit:
+                self.in_transit.discard(el)
+            await self.mlock.release(el.inode, el.blockno)
+            raise
 
         obj_lock_taken = False
         try:
@@ -605,10 +486,9 @@ class BlockCache(object):
                 self.db.execute('INSERT OR REPLACE INTO inode_blocks (block_id, inode, blockno) '
                                 'VALUES(?,?,?)', (block_id, el.inode, el.blockno))
 
-                with lock_released:
-                    self._lock_obj(obj_id)
-                    obj_lock_taken = True
-                    self._queue_upload((el, obj_id))
+                await self.mlock.acquire(obj_id)
+                obj_lock_taken = True
+                await self._queue_upload((el, obj_id))
 
             # There is a block with the same hash
             else:
@@ -621,7 +501,7 @@ class BlockCache(object):
 
                 el.dirty = False
                 self.in_transit.remove(el)
-                self._unlock_entry(el.inode, el.blockno, release_global=True)
+                await self.mlock.release(el.inode, el.blockno)
 
                 if old_block_id == block_id:
                     log.debug('unchanged, block_id=%d', block_id)
@@ -629,25 +509,25 @@ class BlockCache(object):
 
         except:
             self.in_transit.discard(el)
-            with lock_released:
-                self._unlock_entry(el.inode, el.blockno, noerror=True)
-                if obj_lock_taken:
-                    self._unlock_obj(obj_id)
+            await self.mlock.release(el.inode, el.blockno, noerror=True)
+            if obj_lock_taken:
+                await self.mlock.release(obj_id)
             raise
 
         if old_block_id:
-            self._deref_block(old_block_id)
+            await self._deref_block(old_block_id)
         else:
             log.debug('no old block')
 
         return obj_lock_taken
 
 
-    def _queue_upload(self, obj):
+    async def _queue_upload(self, obj):
         '''Put *obj* into upload queue'''
 
         while True:
-            if self.to_upload.put(obj, timeout=5):
+            with trio.move_on_after(5):
+                await self.to_upload[0].send(obj)
                 return
             for t in self.upload_threads:
                 if t.is_alive():
@@ -672,14 +552,11 @@ class BlockCache(object):
             else:
                 raise NoWorkerThreads('no removal threads')
 
-    def _deref_block(self, block_id):
+    async def _deref_block(self, block_id):
         '''Decrease reference count for *block_id*
 
-        If reference counter drops to zero, remove block and propagate
-        to objects table (possibly removing the referenced object
-        as well).
-
-        This method releases the global lock.
+        If reference counter drops to zero, remove block and propagate to
+        objects table (possibly removing the referenced object as well).
         '''
 
         refcount = self.db.get_val('SELECT refcount FROM blocks WHERE id=?', (block_id,))
@@ -707,19 +584,20 @@ class BlockCache(object):
         # the object is no longer in the database.
         log.debug('adding %d to removal queue', obj_id)
 
-        with lock_released:
-            self._lock_obj(obj_id)
-            self._unlock_obj(obj_id)
+        await self.mlock.acquire(obj_id)
+        await self.mlock.release(obj_id)
 
-            if size == -1:
-                # size == -1 indicates that object has not yet been uploaded.
-                # However, since we just acquired a lock on the object, we know
-                # that the upload must have failed. Therefore, trying to remove
-                # this object would just give us another error.
-                return
-
-            self._queue_removal(obj_id)
+        if size == -1:
+            # size == -1 indicates that object has not yet been uploaded.
+            # However, since we just acquired a lock on the object, we know
+            # that the upload must have failed. Therefore, trying to remove
+            # this object would just give us another error.
+            return
 
+        try:
+            self.to_remove.put(obj_id, block=False)
+        except QueueFull:
+            await trio.to_thread.run_sync(self._queue_removal, obj_id)
 
     def transfer_in_progress(self):
         '''Return True if there are any cache entries being uploaded'''
@@ -727,7 +605,10 @@ class BlockCache(object):
         return len(self.in_transit) > 0
 
     def _removal_loop_multi(self):
-        '''Process removal queue'''
+        '''Process removal queue.
+
+        This method runs in a separate thread outside the trio event loop.
+        '''
 
         # This method may look more complicated than necessary, but it ensures
         # that we read as many objects from the queue as we can without
@@ -757,7 +638,10 @@ class BlockCache(object):
                 break
 
     def _removal_loop_simple(self):
-        '''Process removal queue'''
+        '''Process removal queue.
+
+        This method runs in a separate thread outside the trio event loop.
+        '''
 
         while True:
             log.debug('reading from queue..')
@@ -771,21 +655,18 @@ class BlockCache(object):
                     log.warning('Backend lost object s3ql_data_%d' % id_)
                     self.fs.failsafe = True
 
-    @contextmanager
-    def get(self, inode, blockno):
-        """Get file handle for block `blockno` of `inode`
-
-        This method releases the global lock.
-        """
+    @asynccontextmanager
+    async def get(self, inode, blockno):
+        """Get file handle for block `blockno` of `inode`."""
 
         #log.debug('started with %d, %d', inode, blockno)
 
         if self.cache.is_full():
-            self.expire()
+            await self.expire()
 
-        self._lock_entry(inode, blockno, release_global=True)
+        await self.mlock.acquire(inode, blockno)
         try:
-            el = self._get_entry(inode, blockno)
+            el = await self._get_entry(inode, blockno)
             oldsize = el.size
             try:
                 yield el
@@ -794,11 +675,11 @@ class BlockCache(object):
                 # thread has access to a cache entry at any time.
                 self.cache.size += el.size - oldsize
         finally:
-            self._unlock_entry(inode, blockno, release_global=True)
+            await self.mlock.release(inode, blockno)
 
         #log.debug('finished')
 
-    def _get_entry(self, inode, blockno):
+    async def _get_entry(self, inode, blockno):
         '''Get cache entry for `blockno` of `inode`
 
         Assume that cache entry lock has been acquired.
@@ -832,16 +713,18 @@ class BlockCache(object):
                     tmpfh.truncate()
                     shutil.copyfileobj(fh, tmpfh, BUFSIZE)
 
-                with lock_released:
-                    # Lock object. This ensures that we wait until the object
-                    # is uploaded. We don't have to worry about deletion, because
-                    # as long as the current cache entry exists, there will always be
-                    # a reference to the object (and we already have a lock on the
-                    # cache entry).
-                    self._lock_obj(obj_id)
-                    self._unlock_obj(obj_id)
+                # Lock object. This ensures that we wait until the object
+                # is uploaded. We don't have to worry about deletion, because
+                # as long as the current cache entry exists, there will always be
+                # a reference to the object (and we already have a lock on the
+                # cache entry).
+                await self.mlock.acquire(obj_id)
+                await self.mlock.release(obj_id)
+
+                def with_lock_released():
                     with self.backend_pool() as backend:
                         backend.perform_read(do_read, 's3ql_data_%d' % obj_id)
+                await trio.to_thread.run_sync(with_lock_released)
 
                 tmpfh.flush()
                 os.fsync(tmpfh.fileno())
@@ -863,11 +746,8 @@ class BlockCache(object):
 
         return el
 
-    def expire(self):
-        """Perform cache expiry
-
-        This method releases the global lock.
-        """
+    async def expire(self):
+        """Perform cache expiry."""
 
         # Note that we have to make sure that the cache entry is written into
         # the database before we remove it from the cache!
@@ -892,11 +772,11 @@ class BlockCache(object):
                 need_entries -= 1
                 need_size -= el.size
 
-                if self.upload_if_dirty(el): # Releases global lock
+                if await self.upload_if_dirty(el):
                     sth_in_transit = True
                     continue
 
-                self._lock_entry(el.inode, el.blockno, release_global=True)
+                await self.mlock.acquire(el.inode, el.blockno)
                 try:
                     # May have changed while we were waiting for lock
                     if el is not self.cache.get((el.inode, el.blockno), None):
@@ -908,23 +788,21 @@ class BlockCache(object):
                     log.debug('removing %s from cache', el)
                     self.cache.remove((el.inode, el.blockno))
                 finally:
-                    self._unlock_entry(el.inode, el.blockno, release_global=True)
+                    await self.mlock.release(el.inode, el.blockno)
 
             if sth_in_transit:
                 log.debug('waiting for transfer threads..')
-                self.wait() # Releases global lock
+                await self.wait()
 
         log.debug('finished')
 
 
-    def remove(self, inode, start_no, end_no=None):
+    async def remove(self, inode, start_no, end_no=None):
         """Remove blocks for `inode`
 
         If `end_no` is not specified, remove just the `start_no` block.
         Otherwise removes all blocks from `start_no` to, but not including,
          `end_no`.
-
-        This method releases the global lock.
         """
 
         log.debug('started with %d, %d, %s', inode, start_no, end_no)
@@ -940,8 +818,11 @@ class BlockCache(object):
         # waiting for every block to be uploaded only to remove it afterwards.
         for timeout in (0, None):
             for blockno in list(blocknos):
-                if not self._lock_entry(inode, blockno, release_global=True, timeout=timeout):
-                    continue
+                if timeout == 0:
+                    if not self.mlock.acquire_nowait(inode, blockno):
+                        continue
+                else:
+                    await self.mlock.acquire(inode, blockno)
                 blocknos.remove(blockno)
                 try:
                     if (inode, blockno) in self.cache:
@@ -960,10 +841,10 @@ class BlockCache(object):
                                     (inode, blockno))
 
                 finally:
-                    self._unlock_entry(inode, blockno, release_global=True)
+                    await self.mlock.release(inode, blockno)
 
                 # Decrease block refcount
-                self._deref_block(block_id)
+                await self._deref_block(block_id)
 
         log.debug('finished')
 
@@ -977,58 +858,56 @@ class BlockCache(object):
 
         el.flush()
 
-    def start_flush(self):
+    async def start_flush(self, inode=None):
         """Initiate upload of all dirty blocks
 
         When the method returns, all blocks have been registered
         in the database (but the actual uploads may still be
         in progress).
-
-        This method releases the global lock.
         """
 
-        # Need to make copy, since dict() may change while global lock is
-        # released. Look at the comments in CommitThread.run() (mount.py) for an
-        # estimate of the performance impact.
-        for el in list(self.cache.values()):
-            self.upload_if_dirty(el) # Releases global lock
+        # Need to make copy, since dict() may change while uploading.  Look at
+        # the comments in CommitTask.run() (mount.py) for an estimate of the
+        # performance impact.
+        if inode is None:
+            to_flush = list(self.cache.values())
+        else:
+            to_flush = [ x for x in self.cache.values()
+                         if x.inode == inode ]
 
-    def flush(self):
-        """Upload all dirty blocks
+        for el in to_flush:
+            await self.upload_if_dirty(el)
 
-        This method releases the global lock.
-        """
+    async def flush(self):
+        """Upload all dirty blocks."""
 
         log.debug('started')
 
         while True:
             sth_in_transit = False
 
-            # Need to make copy, since dict() may change while global lock is
-            # released. Look at the comments in CommitThread.run() (mount.py)
-            # for an estimate of the performance impact.
+            # Need to make copy, since dict() may change while uploading.  Look
+            # at the comments in CommitTask.run() (mount.py) for an estimate of
+            # the performance impact.
             for el in list(self.cache.values()):
-                if self.upload_if_dirty(el): # Releases global lock
+                if await self.upload_if_dirty(el):
                     sth_in_transit = True
 
             if not sth_in_transit:
                 break
 
             log.debug('waiting for transfer threads..')
-            self.wait() # Releases global lock
+            await self.wait()
 
         log.debug('finished')
 
-    def drop(self):
-        """Drop cache
-
-        This method releases the global lock.
-        """
+    async def drop(self):
+        """Drop cache."""
 
         log.debug('started')
         bak = self.cache.max_entries
         self.cache.max_entries = 0
-        self.expire() # Releases global lock
+        await self.expire()
         self.cache.max_entries = bak
         log.debug('finished')
 
