@@ -21,7 +21,7 @@ import time
 import subprocess
 import errno
 import hashlib
-import llfuse
+import pyfuse3
 import posixpath
 import functools
 import contextlib
@@ -194,7 +194,7 @@ def assert_s3ql_fs(path):
         raise
 
     ctrlfile = os.path.join(path, CTRL_NAME)
-    if not (CTRL_NAME not in llfuse.listdir(path)
+    if not (CTRL_NAME not in pyfuse3.listdir(path)
             and os.path.exists(ctrlfile)):
         raise QuietError('%s is not on an S3QL file system' % path)
 
@@ -324,6 +324,9 @@ def get_backend_factory(options):
             except CorruptedObjectError:
                 raise QuietError('File system revision needs upgrade '
                                  '(or backend data is corrupted)', exitcode=32)
+            except NoSuchObject:
+                raise QuietError('No S3QL file system found at given storage URL.',
+                                 exitcode=18)
 
     return lambda: ComprencBackend(data_pw, compress, options.backend_class(options))
 
@@ -543,14 +546,25 @@ def load_params(cachepath):
         return thaw_basic_mapping(fh.read())
 
 def save_params(cachepath, param):
-    with open(cachepath + '.params', 'wb') as fh:
+    filename = cachepath + '.params'
+    tmpname  = filename + '.tmp'
+    with open(tmpname, 'wb') as fh:
         fh.write(freeze_basic_mapping(param))
-
         # Fsync to make sure that the updated sequence number is committed to
         # disk. Otherwise, a crash immediately after mount could result in both
         # the local and remote metadata appearing to be out of date.
         fh.flush()
         os.fsync(fh.fileno())
 
+    # we need to flush the dirents too.
+    # stackoverflow.com/a/41362774
+    # stackoverflow.com/a/5809073
+    os.rename(tmpname, filename)
+    dirfd = os.open(os.path.dirname(filename), os.O_DIRECTORY)
+    try:
+        os.fsync(dirfd)
+    finally:
+        os.close(dirfd)
+
 def time_ns():
     return int(time.time() * 1e9)
